#!/usr/bin/awk -f
#============================================================================
# @(#) $Header$
#
# NAME		: flat2csv.awk
#
# DESCRIPTION	: Using a description of the fields, it neatly breaks the
#		: file into comma seperated values ready for loading into
#		: a spreadsheet.
#
# USAGE		: awk -f flat2csv.awk FORMAT_FILE INPUT_FILE [INPUT_FILE ...]
#		: The suffix ".csv" will be appended to the filename to
#		: form the output file name.
#
# SEE ALSO	: 
#
# HISTORY
#   1.2  30-May-2004  Tak Tang
#   Now substitutes file suffix instead of simply appending it.
#
#   1.1  ??-Apr-2001  Tak Tang
#   Initial version to help C&CM check NWCS files.
#
#============================================================================


#----------------------------------------------------------------------------
# Set up variables
BEGIN {
  st_skip	= 0		# STATE : Skip file
  st_format	= 1		# STATE : Reading a "format" file
  st_data	= 2		# STATE : Processing a data file
  state		= st_skip	# Processing a format file or a data file?
  format_items	= 0		# Number of known fields
  last_pos	= 0		# Max length of output line (generated)
  seperator	= ","		# Column seperator
  comment	= "#"		# character prefix for comment lines
  doc		= ">"		# character prefix for document lines
  file_suffix	= ".csv"	# output file suffix
  num_records	= 0		# number of records processed
  OUTNAME	= ""		# Output file name
}


#----------------------------------------------------------------------------
# NAME		: store_format_item()
# DESCRIPTION	: Stores a field description into an array for use later
# SYNOPSIS	: store_format_item ( start_position, width, column_name );
# VARIABLES	: "format_items" records the number of entries
#		: "dpos[]" stores the start positions
#		: "dlen[]" stores the field lengths
#		: "dcol[]" stores the column names
#		: "last_pos" records the last character position used
#		: The arrays start at 1.
#----------------------------------------------------------------------------
function store_format_item( startpos, width, endpos, colname) {
  if (startpos != int(startpos)) {
    printf("\aError on line %s ('%s').  Start position is not a number.\n", NR, colname);
    return
  }

  if (width != int(width)) {
    printf("\aError on line %s ('%s').  Field Width is not a number.\n", NR, colname);
    return
  }

  if (endpos != int(endpos)) {
    printf("\aError on line %s ('%s').  End position is not a number.\n", NR, colname);
    return
  }

  if (startpos + width - 1 != endpos) {
    printf("\aError on line %s ('%s').  Start/End Position and width are inconsistent.\n", NR, colname);
    return
  }

  format_items++
  dpos[format_items] = startpos
  dlen[format_items] = width
  dcol[format_items] = colname
  if ((startpos + width - 1) > last_pos) { last_pos = startpos + width - 1 }#
}


#----------------------------------------------------------------------------
# NAME		: process_line()
# DESCRIPTION	: Chop a line up using the field descriptions, and output
#		: each field seperated appropriately.  Fields which contain
#		: a seperator will be enclosed in double quotes.  Any
#		: double quotes in a field containing a seperator will be
#		: replaced by TWO double quotes.
# SYNOPSIS	: process_line ( data_line );
# VARIABLES	: "seperator"
#		: "OUTNAME" -- output file name
#		: The field descriptions "format_items", "dpos[]", "dlen[]"
#----------------------------------------------------------------------------
function process_line( data,   j, d, s ) {
  s = seperator
  for (j=1; j<=format_items; j++) {	# for each field we know about
    if (j == format_items) { s="\n"; }
    d = substr(data, dpos[j], dlen[j]) # extract field from line
    sub("^  *", "", d);		# strip leading spaces
    sub("  *$", "", d);		# strip trailing spaces
    if (index(d,"\"")>0 || index(d, seperator)> 0) {
      gsub("\"", "\"\"", d);	# replace double quotes with two double quotes
      d="\""d"\"";		# enclose value in quotes
    }
    printf("%s"s, d ) >OUTNAME	# write this field out
  } # for each format item
}


function process_line_faster( data,   j ) {
  for (j=1; j<format_items; j++) {	# for each field we know about
    printf("%s"seperator, substr(data, dpos[j], dlen[j]) ) >OUTNAME	# write this field out
  } # for each format item
  printf("%s\n", substr(data, dpos[j], dlen[j]) ) >OUTNAME	# write this field out
}


#----------------------------------------------------------------------------
#
function print_Column_Headers(    j ) {
  for (j=1; j<format_items; j++) {
    printf("%s"seperator,  dcol[j] ) >OUTNAME
  } # for each format item
  printf("%s\n",  dcol[j] ) >OUTNAME
}


#----------------------------------------------------------------------------
#
function print_underline(    d ) {
  d=sprintf("%"last_pos"s", " ");
  gsub(" ", "-", d);
  process_line( sprintf("%s",d) );
}


#----------------------------------------------------------------------------
# First line
1==FNR {	# Is this the first line of the file?
  if (OUTNAME != "") {
    close(OUTNAME);
    OUTNAME = "";
  }

  if ("FORMAT"==$1) {	# Is this a "FORMAT" file?
    state = st_format	# now reading a format file
    TABLE_NAME = $2	# make a note of the type of file this is
    format_items = 0	# No field descriptions loaded yet
    printf("Loading %s record description from %s\n", TABLE_NAME, FILENAME);
    next
  } # end if this is a format file
  else { # probably a data file
    if (format_items>0) {	# do we have any format items to apply to this data file
      state = st_data		# This must be a data file
      OUTNAME = FILENAME	# build output file name
      sub(/(\.[^.]+)?$/, file_suffix, OUTNAME); # substitute correct suffix
      printf("\nProcessing file '%s' -> '%s', using %s (%s fields)\n", FILENAME, OUTNAME, TABLE_NAME, format_items);
      print_Column_Headers();
      print_underline();
    } else {
      printf("ERROR!  Found datafile '%s', but we do not have any fields!\n", FILENAME);
      state = st_skip
    }
  } # end if this is not a format file
} # end if first line of a file


#----------------------------------------------------------------------------
# A data line
state==st_data { process_line($0) }


#----------------------------------------------------------------------------
# A table format descriptor
state==st_format && substr($0,1,1) == comment { next; }
state==st_format && substr($0,1,2) == doc" " { print substr($0,3); next; }
state==st_format && substr($0,1,1) == doc { print substr($0,2); next; }
state==st_format && length($0)>0 { store_format_item($2, $1, $3, $4); next; }


#============================================================================
# END OF FILE
